@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
+# Copyright (c) 2018-2022 Micro Focus or one of its affiliates.
 # Copyright (c) 2018 Uber Technologies, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -64,6 +64,7 @@ from ..vertica.log import VerticaLogging
 DEFAULT_HOST = 'localhost'
 DEFAULT_PORT = 5433
 DEFAULT_PASSWORD = ''
+DEFAULT_AUTOCOMMIT = False
 DEFAULT_BACKUP_SERVER_NODE = []
 DEFAULT_KRB_SERVICE_NAME = 'vertica'
 DEFAULT_LOG_LEVEL = logging.WARNING
@@ -109,7 +110,8 @@ def parse_dsn(dsn):
             continue
         elif key == 'backup_server_node':
             continue
-        elif key in ('connection_load_balance', 'use_prepared_statements', 'ssl'):
+        elif key in ('connection_load_balance', 'use_prepared_statements',
+                     'disable_copy_local', 'ssl', 'autocommit'):
             lower = value.lower()
             if lower in ('true', 'on', '1'):
                 result[key] = True
@@ -277,6 +279,7 @@ class Connection(object):
                 raise KeyError(msg)
         self.options.setdefault('database', self.options['user'])
         self.options.setdefault('password', DEFAULT_PASSWORD)
+        self.options.setdefault('autocommit', DEFAULT_AUTOCOMMIT)
         self.options.setdefault('session_label', _generate_session_label())
         self.options.setdefault('backup_server_node', DEFAULT_BACKUP_SERVER_NODE)
         self.options.setdefault('kerberos_service_name', DEFAULT_KRB_SERVICE_NAME)
@@ -297,10 +300,20 @@ class Connection(object):
         self._logger.debug('Connection prepared statements is {}'.format(
                      'enabled' if self.options['use_prepared_statements'] else 'disabled'))
 
+        # knob for disabling COPY LOCAL operations
+        self.options.setdefault('disable_copy_local', False)
+        self._logger.debug('COPY LOCAL operation is {}'.format(
+                     'disabled' if self.options['disable_copy_local'] else 'enabled'))
+
         self._logger.info('Connecting as user "{}" to database "{}" on host "{}" with port {}'.format(
                      self.options['user'], self.options['database'],
                      self.options['host'], self.options['port']))
         self.startup_connection()
+
+        # Initially, for a new session, autocommit is off
+        if self.options['autocommit']:
+            self.autocommit = True
+
         self._logger.info('Connection is ready')
 
     #############################################
@@ -310,20 +323,7 @@ class Connection(object):
         return self
 
     def __exit__(self, type_, value, traceback):
-        try:
-            # get the transaction status
-            if not self.closed():
-                self.cursor().flush_to_query_ready()
-            # if there's no outstanding transaction, we can simply close the connection
-            if self.transaction_status in (None, 'in_transaction'):
-                return
-
-            if type_ is not None:
-                self.rollback()
-            else:
-                self.commit()
-        finally:
-            self.close()
+        self.close()
 
     #############################################
     # dbapi methods
@@ -363,6 +363,21 @@ class Connection(object):
     #############################################
     # non-dbapi methods
     #############################################
+    @property
+    def autocommit(self):
+        """Read the connection's AUTOCOMMIT setting from cache"""
+        return self.parameters.get('auto_commit', 'off') == 'on'
+
+    @autocommit.setter
+    def autocommit(self, value):
+        """Change the connection's AUTOCOMMIT setting"""
+        if self.autocommit is value:
+            return
+        val = 'on' if value else 'off'
+        cur = self.cursor()
+        cur.execute('SET SESSION AUTOCOMMIT TO {}'.format(val), use_prepared_statements=False)
+        cur.fetchall()   # check for errors and update the cache
+
     def cancel(self):
         """Cancel the current database operation. This can be called from a
            different thread than the one currently executing a database operation.
@@ -553,16 +568,20 @@ class Connection(object):
         self._logger.debug('=> %s', message)
         try:
             for data in message.fetch_message():
-                try:
-                    vsocket.sendall(data)
-                except Exception:
-                    self._logger.error("couldn't send message")
-                    raise
-
+                size = 8192 # Max msg size, consistent with how the server works
+                pos = 0
+                while pos < len(data):
+                    sent = vsocket.send(data[pos : pos + size])
+                    if sent == 0:
+                        raise errors.ConnectionError("Couldn't send message: Socket connection broken")
+                    pos += sent
         except Exception as e:
             self.close_socket()
             self._logger.error(str(e))
-            raise
+            if isinstance(e, IOError):
+                raise_from(errors.ConnectionError(str(e)), e)
+            else:
+                raise
 
     def close_socket(self):
         try:
@@ -594,6 +613,15 @@ class Connection(object):
             else:
                 self._logger.warning(message.error_message())
 
+    def read_string(self):
+        s = bytearray()
+        while True:
+            char = self.read_bytes(1)
+            if char == b'\x00':
+                break
+            s.extend(char)
+        return s
+
     def read_message(self):
         while True:
             try:
@@ -601,7 +629,27 @@ class Connection(object):
                 size = unpack('!I', self.read_bytes(4))[0]
                 if size < 4:
                     raise errors.MessageError("Bad message size: {0}".format(size))
-                message = BackendMessage.from_type(type_, self.read_bytes(size - 4))
+                if type_ == messages.WriteFile.message_id:
+                    # The whole WriteFile message may not be read at here.
+                    # Instead, only the file name and file length is read.
+                    # This is because the message could be too large to read all at once.
+                    f = self.read_string()
+                    filename = f.decode('utf-8')
+                    file_length = unpack('!I', self.read_bytes(4))[0]
+                    size -= 4 + len(f) + 1 + 4
+                    if size != file_length:
+                        raise errors.MessageError("Bad message size: {0}".format(size))
+                    if filename == '':
+                        # If there is no filename, then this is really RETURNREJECTED data, not a rejected file
+                        if file_length % 8 != 0:
+                            raise errors.MessageError("Bad RETURNREJECTED data size: {0}".format(file_length))
+                        data = self.read_bytes(file_length)
+                        message = messages.WriteFile(filename, file_length, data)
+                    else:
+                        # The rest of the message is read later with write_to_disk()
+                        message = messages.WriteFile(filename, file_length)
+                else:
+                    message = BackendMessage.from_type(type_, self.read_bytes(size - 4))
                 self._logger.debug('<= %s', message)
                 self.handle_asynchronous_message(message)
                 # handle transaction status
@@ -637,13 +685,22 @@ class Connection(object):
             raise errors.MessageError(msg)
 
     def read_bytes(self, n):
-        results = bytes()
-        while len(results) < n:
-            bytes_ = self._socket().recv(n - len(results))
-            if not bytes_:
+        if n == 1:
+            result = self._socket().recv(1)
+            if not result:
                 raise errors.ConnectionError("Connection closed by Vertica")
-            results += bytes_
-        return results
+            return result
+        else:
+            buf = bytearray(n)
+            view = memoryview(buf)
+            to_read = n
+            while to_read > 0:
+                received = self._socket().recv_into(view, to_read)
+                if received == 0:
+                    raise errors.ConnectionError("Connection closed by Vertica")
+                view = view[received:]
+                to_read -= received
+            return buf
 
     def send_GSS_response_and_receive_challenge(self, response):       
         # Send the GSS response data to the vertica server
